'use strict';

const util = require('util');
const Module = require('module');
const promises = require('fs/promises');
const upath = require('upath');
const chalk = require('chalk');
const consola = require('consola');
const rimraf = require('rimraf');
const defu = require('defu');
const prettyBytes = require('pretty-bytes');
const jiti = require('jiti');
const rollup = require('rollup');
const commonjs = require('@rollup/plugin-commonjs');
const nodeResolve = require('@rollup/plugin-node-resolve');
const alias = require('@rollup/plugin-alias');
const esbuild = require('rollup-plugin-esbuild');
const dts = require('rollup-plugin-dts');
const untyped = require('untyped');
const scule = require('scule');
const mkdist = require('mkdist');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

function _interopNamespace(e) {
  if (e && e.__esModule) return e;
  var n = Object.create(null);
  if (e) {
    Object.keys(e).forEach(function (k) {
      if (k !== 'default') {
        var d = Object.getOwnPropertyDescriptor(e, k);
        Object.defineProperty(n, k, d.get ? d : {
          enumerable: true,
          get: function () {
            return e[k];
          }
        });
      }
    });
  }
  n['default'] = e;
  return Object.freeze(n);
}

const Module__default = /*#__PURE__*/_interopDefaultLegacy(Module);
const chalk__default = /*#__PURE__*/_interopDefaultLegacy(chalk);
const consola__default = /*#__PURE__*/_interopDefaultLegacy(consola);
const rimraf__default = /*#__PURE__*/_interopDefaultLegacy(rimraf);
const defu__default = /*#__PURE__*/_interopDefaultLegacy(defu);
const prettyBytes__default = /*#__PURE__*/_interopDefaultLegacy(prettyBytes);
const jiti__default = /*#__PURE__*/_interopDefaultLegacy(jiti);
const commonjs__default = /*#__PURE__*/_interopDefaultLegacy(commonjs);
const nodeResolve__default = /*#__PURE__*/_interopDefaultLegacy(nodeResolve);
const alias__default = /*#__PURE__*/_interopDefaultLegacy(alias);
const esbuild__default = /*#__PURE__*/_interopDefaultLegacy(esbuild);
const dts__default = /*#__PURE__*/_interopDefaultLegacy(dts);

function validateDependencies(ctx) {
  const usedDependencies = new Set();
  const unusedDependencies = new Set(Object.keys(ctx.pkg.dependencies || {}));
  const implicitDependnecies = new Set();
  for (const id of ctx.usedImports) {
    unusedDependencies.delete(id);
    usedDependencies.add(id);
  }
  if (Array.isArray(ctx.dependencies)) {
    for (const id of ctx.dependencies) {
      unusedDependencies.delete(id);
    }
  }
  for (const id of usedDependencies) {
    if (!ctx.externals.includes(id) && !id.startsWith("chunks/") && !ctx.externals.includes(id.split("/")[0])) {
      implicitDependnecies.add(id);
    }
  }
  if (unusedDependencies.size) {
    consola__default['default'].warn("Potential unused dependencies found:", Array.from(unusedDependencies).map((id) => chalk__default['default'].cyan(id)).join(", "));
  }
  if (implicitDependnecies.size) {
    consola__default['default'].warn("Potential implicit dependencies found:", Array.from(implicitDependnecies).map((id) => chalk__default['default'].cyan(id)).join(", "));
  }
}

async function rollupBuild(ctx) {
  if (ctx.stub) {
    for (const entry of ctx.entries.filter((entry2) => entry2.builder === "rollup")) {
      const input = upath.resolve(ctx.rootDir, entry.input);
      const output = upath.resolve(ctx.rootDir, ctx.outDir, entry.name);
      await promises.writeFile(output + ".js", `module.exports = require('jiti')()('${input}')`);
      await promises.writeFile(output + ".mjs", `export * from '${input}'`);
      await promises.writeFile(output + ".d.ts", `export * from '${input}'`);
    }
    return;
  }
  const rollupOptions = getRollupOptions(ctx);
  if (!Object.keys(rollupOptions.input).length) {
    return;
  }
  console.log(ctx.externals);
  const buildResult = await rollup.rollup(rollupOptions);
  const allOutputOptions = rollupOptions.output;
  for (const outputOptions of allOutputOptions) {
    const {output} = await buildResult.write(outputOptions);
    for (const entry of output.filter((e) => e.type === "chunk")) {
      for (const id of entry.imports) {
        ctx.usedImports.add(id);
      }
      if (entry.isEntry) {
        ctx.buildEntries.push({
          path: upath.relative(ctx.rootDir, upath.resolve(outputOptions.dir, entry.fileName)),
          bytes: entry.code.length * 4,
          exports: entry.exports
        });
      }
    }
  }
  rollupOptions.plugins = rollupOptions.plugins || [];
  rollupOptions.plugins.push(dts__default['default']());
  const typesBuild = await rollup.rollup(rollupOptions);
  await typesBuild.write({
    dir: upath.resolve(ctx.rootDir, ctx.outDir),
    format: "esm"
  });
}
function getRollupOptions(ctx) {
  const extensions = [".ts", ".mjs", ".js", ".json"];
  return {
    context: ctx.rootDir,
    input: Object.fromEntries(ctx.entries.filter((entry) => entry.builder === "rollup").map((entry) => [entry.name, upath.resolve(ctx.rootDir, entry.input)])),
    output: [
      {
        dir: upath.resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].js",
        chunkFileNames: "chunks/[name].js",
        format: "cjs",
        exports: "auto",
        preferConst: true
      },
      {
        dir: upath.resolve(ctx.rootDir, ctx.outDir),
        entryFileNames: "[name].mjs",
        chunkFileNames: "chunks/[name].mjs",
        format: "esm",
        exports: "auto",
        preferConst: true
      }
    ],
    external(id) {
      if (id[0] === "." || id[0] === "/" || id.includes("src/")) {
        return false;
      }
      const isExplicitExternal = !!ctx.externals.find((ext) => id.includes(ext));
      if (!isExplicitExternal) {
        consola__default['default'].warn(`Inlining external ${id}`);
      }
      return isExplicitExternal;
    },
    onwarn(warning, rollupWarn) {
      if (!warning.code || !["CIRCULAR_DEPENDENCY"].includes(warning.code)) {
        rollupWarn(warning);
      }
    },
    plugins: [
      alias__default['default']({}),
      nodeResolve__default['default']({
        extensions
      }),
      esbuild__default['default']({
        target: "node12",
        loaders: {
          ".json": "json"
        }
      }),
      commonjs__default['default']({
        extensions
      })
    ]
  };
}

async function typesBuild(ctx) {
  for (const entry of ctx.entries.filter((entry2) => entry2.builder === "untyped")) {
    const srcConfig = await Promise.resolve().then(function () { return /*#__PURE__*/_interopNamespace(require(upath.resolve(ctx.rootDir, entry.input))); }).then((r) => r.default);
    const genDir = upath.resolve(ctx.rootDir, ctx.genDir, entry.name);
    const defaults = entry.defaults || {};
    const schema = untyped.resolveSchema(srcConfig, defaults);
    await promises.mkdir(upath.dirname(genDir)).catch(() => {
    });
    await promises.mkdir(genDir).catch(() => {
    });
    await promises.writeFile(upath.resolve(genDir, `${entry.name}.md`), untyped.generateMarkdown(schema));
    await promises.writeFile(upath.resolve(genDir, `${entry.name}.schema.json`), JSON.stringify(schema, null, 2));
    await promises.writeFile(upath.resolve(genDir, `${entry.name}.defaults.json`), JSON.stringify(defaults, null, 2));
    await promises.writeFile(upath.resolve(genDir, `${entry.name}.d.ts`), "export " + untyped.generateTypes(schema, scule.pascalCase(entry.name + "-schema")));
  }
}

async function mkdistBuild(ctx) {
  for (const entry of ctx.entries.filter((e) => e.builder === "mkdist")) {
    if (ctx.stub) {
      const srcDir = upath.resolve(ctx.rootDir, entry.input);
      const outDir = upath.resolve(ctx.rootDir, ctx.outDir, entry.name);
      await promises.unlink(outDir).catch(() => {
      });
      await promises.symlink(srcDir, outDir);
    } else {
      const {writtenFiles} = await mkdist.mkdist({
        rootDir: ctx.rootDir,
        srcDir: entry.input,
        distDir: upath.join(ctx.outDir, entry.name),
        format: entry.format,
        cleanDist: false
      });
      ctx.buildEntries.push({
        path: upath.join(ctx.outDir, entry.name),
        chunks: [`${writtenFiles.length} files`]
      });
    }
  }
}

async function build(rootDir, stub) {
  rootDir = upath.resolve(process.cwd(), rootDir || ".");
  const _require = jiti__default['default'](rootDir);
  let buildConfigFile;
  try {
    buildConfigFile = _require.resolve("./build.config");
  } catch (e) {
  }
  const buildConfig = buildConfigFile ? _require("./build.config").default : {};
  const pkg = _require("./package.json");
  const ctx = defu__default['default'](buildConfig, {
    pkg,
    rootDir,
    entries: [],
    dependencies: [],
    externals: [...Module__default['default'].builtinModules],
    srcDir: "src",
    outDir: "dist",
    genDir: ".gen",
    untyped: void 0,
    clean: true,
    stub,
    buildEntries: [],
    usedImports: new Set()
  });
  ctx.entries = ctx.entries.map((entry) => typeof entry === "string" ? {input: entry} : entry);
  for (const entry of ctx.entries) {
    if (typeof entry.name !== "string") {
      entry.name = upath.basename(entry.input);
    }
    if (!entry.input) {
      throw new Error("Missing entry input: " + dumpObject(entry));
    }
    if (!entry.builder) {
      entry.builder = entry.input.endsWith("/") ? "mkdist" : "rollup";
    }
  }
  if (pkg.dependencies) {
    ctx.externals.push(...Object.keys(pkg.dependencies));
  }
  consola__default['default'].info(chalk__default['default'].cyan(`Building ${pkg.name}`));
  if (process.env.DEBUG) {
    consola__default['default'].info(`${chalk__default['default'].bold("Root dir:")} ${ctx.rootDir}
  ${chalk__default['default'].bold("Entries:")}
  ${ctx.entries.map((entry) => "  " + dumpObject(entry)).join("\n  ")}
`);
  }
  if (ctx.clean) {
    const outDir = upath.resolve(ctx.rootDir, ctx.outDir);
    await promises.unlink(outDir).catch(() => {
    });
    await util.promisify(rimraf__default['default'])(outDir);
    await promises.mkdir(outDir).catch(() => {
    });
  }
  await typesBuild(ctx);
  await mkdistBuild(ctx);
  await rollupBuild(ctx);
  consola__default['default'].success(chalk__default['default'].green("Build succeed for " + pkg.name));
  for (const entry of ctx.buildEntries) {
    consola__default['default'].log(`  ${chalk__default['default'].bold(entry.path)} (` + [
      entry.bytes && `size: ${chalk__default['default'].cyan(prettyBytes__default['default'](entry.bytes))}`,
      entry.exports && `exports: ${chalk__default['default'].gray(entry.exports.join(", "))}`,
      entry.chunks && `chunks: ${chalk__default['default'].gray(entry.chunks.join(", "))}`
    ].filter(Boolean).join(", ") + ")");
  }
  validateDependencies(ctx);
  consola__default['default'].log("");
}
function dumpObject(obj) {
  return "{ " + Object.keys(obj).map((key) => `${key}: ${JSON.stringify(obj[key])}`).join(", ") + " }";
}

exports.build = build;
